# torchcell/datasets/genome.py
# [[torchcell.datasets.genome]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/genome.py
# Test file: torchcell/datasets/test_genome.py

from abc import ABC, abstractmethod

from torch_geometric.data import Data, InMemoryDataset, download_url


class GenomeDataset(InMemoryDataset, ABC):
    pass

    def __add__():
        pass


class SCerevisiaeS88C(GenomeDataset):
    pass


if __name__ == "__main__":
    pass
