# torchcell/__version__.py
# [[torchcell.__version__]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/__version__.py
# Test file: tests/torchcell/test___version__.py

__version__ = "0.1.17"  # Linked to semantic-release in toml
