# torchcell/ontology/tc_ontology.py
# [[torchcell.ontology.tc_ontology]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/ontology/tc_ontology.py
# Test file: torchcell/ontology/test_tc_ontology.py
